@@ -1,7 +1,10 @@
 import torch
 from torch import nn, einsum
 import torch.nn.functional as F
+from torch.cuda.amp import autocast
+
 from einops import rearrange, repeat
+from contextlib import contextmanager
 
 def exists(val):
     return val is not None
@@ -9,6 +12,10 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
+@contextmanager
+def null_context():
+    yield
+
 def l2norm(t):
     return F.normalize(t, p = 2, dim = -1)
 
@@ -263,7 +270,9 @@ class VectorQuantize(nn.Module):
         commitment_weight = None,
         commitment = 1., # deprecate in next version, turn off by default
         orthogonal_reg_weight = 0.,
-        orthogonal_reg_active_codes_only = False
+        orthogonal_reg_active_codes_only = False,
+        orthogonal_reg_max_codes = None,
+        quantize_full_precision = False
     ):
         super().__init__()
         n_embed = default(n_embed, codebook_size)
@@ -280,6 +289,7 @@ class VectorQuantize(nn.Module):
 
         self.orthogonal_reg_weight = orthogonal_reg_weight
         self.orthogonal_reg_active_codes_only = orthogonal_reg_active_codes_only
+        self.orthogonal_reg_max_codes = orthogonal_reg_max_codes
 
         codebook_class = EuclideanCodebook if not use_cosine_sim \
                          else CosineSimCodebook
@@ -299,6 +309,9 @@ class VectorQuantize(nn.Module):
         self.accept_image_fmap = accept_image_fmap
         self.channel_last = channel_last
 
+        # when quantizing, whether to turn off mixed precision
+        self.quantize_full_precision = quantize_full_precision
+
     @property
     def codebook(self):
         return self._codebook.embed
@@ -317,7 +330,10 @@ class VectorQuantize(nn.Module):
 
         x = self.project_in(x)
 
-        quantize, embed_ind = self._codebook(x)
+        context = null_context() if not self.quantize_full_precision else autocast(enabled = False)
+
+        with context:
+            quantize, embed_ind = self._codebook(x)
 
         if self.training:
             quantize = x + (quantize - x).detach()
@@ -335,11 +351,14 @@ class VectorQuantize(nn.Module):
                 if self.orthogonal_reg_active_codes_only:
                     # only calculate orthogonal loss for the activated codes for this batch
                     unique_code_ids = torch.unique(embed_ind)
-                    codebook = self.codebook[unique_code_ids]
-                else:
-                    codebook = self.codebook
+                    codebook = codebook[unique_code_ids]
 
-                orthogonal_reg_loss = orthgonal_loss_fn(self.codebook)
+                num_codes = codebook.shape[0]
+                if exists(self.orthogonal_reg_max_codes) and num_codes > self.orthogonal_reg_max_codes:
+                    rand_ids = torch.randperm(num_codes, device = device)[:self.orthogonal_reg_max_codes]
+                    codebook = codebook[rand_ids]
+
+                orthogonal_reg_loss = orthgonal_loss_fn(codebook)
                 loss = loss + orthogonal_reg_loss * self.orthogonal_reg_weight
 
         quantize = self.project_out(quantize)
